@@ -34,6 +34,7 @@ along with BTFS.  If not, see <http://www.gnu.org/licenses/>.
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wsign-conversion"
 #pragma GCC diagnostic ignored "-Wconversion"
+#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
 #include <libtorrent/torrent_info.hpp>
 #include <libtorrent/session.hpp>
 #include <libtorrent/alert.hpp>
@@ -98,15 +99,9 @@ jump(int piece, int size) {
 
 	cursor = tail;
 
-	int pl = ti->piece_length();
-
-	for (int b = 0; b < 16 * pl; b += pl) {
+	for (int i = 0; i < 16; i++) {
 		handle.piece_priority(tail++, 7);
 	}
-
-	for (int o = (tail - piece) * pl; o < size + pl - 1; o += pl) {
-		handle.piece_priority(tail++, 1);
-	}
 }
 
 static void
@@ -139,6 +134,13 @@ Read::Read(char *buf, int index, off_t offset, size_t size) {
 	}
 }
 
+void Read::fail(int piece) {
+	for (parts_iter i = parts.begin(); i != parts.end(); ++i) {
+		if (i->part.piece == piece && !i->filled)
+			failed = true;
+	}
+}
+
 void Read::copy(int piece, char *buffer, int size) {
 	for (parts_iter i = parts.begin(); i != parts.end(); ++i) {
 		if (i->part.piece == piece && !i->filled)
@@ -183,11 +185,14 @@ int Read::read() {
 	// Move sliding window to first piece to serve this request
 	jump(parts.front().part.piece, size());
 
-	while (!finished())
+	while (!finished() && !failed)
 		// Wait for any piece to downloaded
 		pthread_cond_wait(&signal_cond, &lock);
 
-	return size();
+	if (failed)
+		return -EIO;
+	else
+		return size();
 }
 
 static void
@@ -200,9 +205,6 @@ setup() {
 		handle.pause();
 
 	for (int i = 0; i < ti->num_files(); ++i) {
-		// Initially, don't download anything
-		handle.file_priority(i, 0);
-
 		std::string parent("");
 
 #if LIBTORRENT_VERSION_NUM < 10100
@@ -211,6 +213,9 @@ setup() {
 		char *p = strdup(ti->files().file_path(i).c_str());
 #endif
 
+		if (!p)
+			continue;
+
 		for (char *x = strtok(p, "/"); x; x = strtok(NULL, "/")) {
 			if (strlen(x) <= 0)
 				continue;
@@ -239,12 +244,21 @@ setup() {
 
 static void
 handle_read_piece_alert(libtorrent::read_piece_alert *a, Log *log) {
-	printf("%s: piece %d size %d\n", __func__, a->piece, a->size);
+	printf("%s: piece %d size %d\n", __func__, static_cast<int>(a->piece),
+		a->size);
 
 	pthread_mutex_lock(&lock);
 
-	for (reads_iter i = reads.begin(); i != reads.end(); ++i) {
-		(*i)->copy(a->piece, a->buffer.get(), a->size);
+	if (a->ec) {
+		*log << a->message() << std::endl;
+
+		for (reads_iter i = reads.begin(); i != reads.end(); ++i) {
+			(*i)->fail(a->piece);
+		}
+	} else {
+		for (reads_iter i = reads.begin(); i != reads.end(); ++i) {
+			(*i)->copy(a->piece, a->buffer.get(), a->size);
+		}
 	}
 
 	pthread_mutex_unlock(&lock);
@@ -255,7 +269,7 @@ handle_read_piece_alert(libtorrent::read_piece_alert *a, Log *log) {
 
 static void
 handle_piece_finished_alert(libtorrent::piece_finished_alert *a, Log *log) {
-	printf("%s: %d\n", __func__, a->piece_index);
+	printf("%s: %d\n", __func__, static_cast<int>(a->piece_index));
 
 	pthread_mutex_lock(&lock);
 
@@ -271,8 +285,6 @@ handle_piece_finished_alert(libtorrent::piece_finished_alert *a, Log *log) {
 
 static void
 handle_torrent_added_alert(libtorrent::torrent_added_alert *a, Log *log) {
-	//printf("%s()\n", __func__);
-
 	pthread_mutex_lock(&lock);
 
 	handle = a->handle;
@@ -286,8 +298,6 @@ handle_torrent_added_alert(libtorrent::torrent_added_alert *a, Log *log) {
 static void
 handle_metadata_received_alert(libtorrent::metadata_received_alert *a,
 		Log *log) {
-	//printf("%s\n", __func__);
-
 	pthread_mutex_lock(&lock);
 
 	handle = a->handle;
@@ -319,14 +329,18 @@ handle_alert(libtorrent::alert *a, Log *log) {
 		handle_torrent_added_alert(
 			(libtorrent::torrent_added_alert *) a, log);
 		break;
+	case libtorrent::dht_bootstrap_alert::alert_type:
+		*log << a->message() << std::endl;
+		// Force DHT announce because libtorrent won't by itself
+		handle.force_dht_announce();
+		break;
+	case libtorrent::dht_announce_alert::alert_type:
+	case libtorrent::dht_reply_alert::alert_type:
 	case libtorrent::metadata_failed_alert::alert_type:
 	case libtorrent::tracker_announce_alert::alert_type:
 	case libtorrent::tracker_reply_alert::alert_type:
 	case libtorrent::tracker_warning_alert::alert_type:
 	case libtorrent::tracker_error_alert::alert_type:
-	case libtorrent::dht_bootstrap_alert::alert_type:
-	case libtorrent::dht_announce_alert::alert_type:
-	case libtorrent::dht_reply_alert::alert_type:
 	case libtorrent::lsd_peer_alert::alert_type:
 		*log << a->message() << std::endl;
 		break;
@@ -390,6 +404,11 @@ alert_queue_loop(void *data) {
 	return NULL;
 }
 
+static bool
+is_root(const char *path) {
+	return strcmp(path, "/") == 0;
+}
+
 static bool
 is_dir(const char *path) {
 	return dirs.find(path) != dirs.end();
@@ -402,7 +421,7 @@ is_file(const char *path) {
 
 static int
 btfs_getattr(const char *path, struct stat *stbuf) {
-	if (!is_dir(path) && !is_file(path) && strcmp(path, "/") != 0)
+	if (!is_dir(path) && !is_file(path) && !is_root(path))
 		return -ENOENT;
 
 	pthread_mutex_lock(&lock);
@@ -413,7 +432,7 @@ btfs_getattr(const char *path, struct stat *stbuf) {
 	stbuf->st_gid = getgid();
 	stbuf->st_mtime = time_of_mount;
 
-	if (strcmp(path, "/") == 0 || is_dir(path)) {
+	if (is_root(path) || is_dir(path)) {
 		stbuf->st_mode = S_IFDIR | 0755;
 	} else {
 		auto ti = handle.torrent_file();
@@ -447,7 +466,7 @@ btfs_getattr(const char *path, struct stat *stbuf) {
 static int
 btfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 		off_t offset, struct fuse_file_info *fi) {
-	if (!is_dir(path) && !is_file(path) && strcmp(path, "/") != 0)
+	if (!is_dir(path) && !is_file(path) && !is_root(path))
 		return -ENOENT;
 
 	if (is_file(path))
@@ -485,8 +504,6 @@ btfs_open(const char *path, struct fuse_file_info *fi) {
 static int
 btfs_read(const char *path, char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi) {
-	//printf("%s: %s %lu %ld\n", __func__, path, size, offset);
-
 	if (!is_dir(path) && !is_file(path))
 		return -ENOENT;
 
@@ -543,7 +560,11 @@ btfs_init(struct fuse_conn_info *conn) {
 	libtorrent::add_torrent_params *p = (libtorrent::add_torrent_params *)
 		fuse_get_context()->private_data;
 
+#if LIBTORRENT_VERSION_NUM < 10200
 	int flags =
+#else
+	libtorrent::session_flags_t flags =
+#endif
 		libtorrent::session::add_default_plugins |
 		libtorrent::session::start_default_features;
 
@@ -576,6 +597,7 @@ btfs_init(struct fuse_conn_info *conn) {
 
 	libtorrent::session_settings se = session->settings();
 
+	se.request_timeout = 10;
 	se.strict_end_game_mode = false;
 	se.announce_to_all_trackers = true;
 	se.announce_to_all_tiers = true;
@@ -613,6 +635,7 @@ btfs_init(struct fuse_conn_info *conn) {
 		"dht.transmissionbt.com:6881");
 #endif
 
+	pack.set_int(pack.request_timeout, 10);
 	pack.set_str(pack.listen_interfaces, interfaces.str());
 	pack.set_bool(pack.strict_end_game_mode, false);
 	pack.set_bool(pack.announce_to_all_trackers, true);
@@ -651,14 +674,91 @@ btfs_destroy(void *user_data) {
 	pthread_cancel(alert_thread);
 	pthread_join(alert_thread, NULL);
 
-	session->remove_torrent(handle,
-		params.keep ? 0 : libtorrent::session::delete_files);
+#if LIBTORRENT_VERSION_NUM < 10200
+	int flags = 0;
+#else
+	libtorrent::remove_flags_t flags = {};
+#endif
+
+	if (!params.keep)
+		flags |= libtorrent::session::delete_files;
+
+	session->remove_torrent(handle, flags);
 
 	delete session;
 
 	pthread_mutex_unlock(&lock);
 }
 
+static int
+btfs_listxattr(const char *path, char *data, size_t len) {
+	const char *xattrs = NULL;
+	int xattrslen = 0;
+
+	if (is_root(path)) {
+		xattrs = XATTR_IS_BTFS "\0" XATTR_IS_BTFS_ROOT;
+		xattrslen = sizeof (XATTR_IS_BTFS "\0" XATTR_IS_BTFS_ROOT);
+	} else if (is_dir(path)) {
+		xattrs = XATTR_IS_BTFS;
+		xattrslen = sizeof (XATTR_IS_BTFS);
+	} else if (is_file(path)) {
+		xattrs = XATTR_IS_BTFS "\0" XATTR_FILE_INDEX;
+		xattrslen = sizeof (XATTR_IS_BTFS "\0" XATTR_FILE_INDEX);
+	} else {
+		return -ENOENT;
+	}
+
+	// The minimum required length
+	if (len == 0)
+		return xattrslen;
+
+	if (len < (size_t) xattrslen)
+		return -ERANGE;
+
+	memcpy(data, xattrs, (size_t) xattrslen);
+
+	return xattrslen;
+}
+
+#ifdef __APPLE__
+static int
+btfs_getxattr(const char *path, const char *key, char *value, size_t len,
+		uint32_t position) {
+#else
+static int
+btfs_getxattr(const char *path, const char *key, char *value, size_t len) {
+	uint32_t position = 0;
+#endif
+	char xattr[16];
+	int xattrlen = 0;
+
+	std::string k(key);
+
+	if (is_file(path) && k == XATTR_FILE_INDEX) {
+		xattrlen = snprintf(xattr, sizeof (xattr), "%d", files[path]);
+	} else if (is_root(path) && k == XATTR_IS_BTFS_ROOT) {
+		xattrlen = 0;
+	} else if (k == XATTR_IS_BTFS) {
+		xattrlen = 0;
+	} else {
+		return -ENODATA;
+	}
+
+	// The minimum required length
+	if (len == 0)
+		return xattrlen;
+
+	if (position >= (uint32_t) xattrlen)
+		return 0;
+
+	if (len < (size_t) xattrlen - position)
+		return -ERANGE;
+
+	memcpy(value, xattr + position, (size_t) xattrlen - position);
+
+	return xattrlen - (int) position;
+}
+
 static bool
 populate_target(std::string& target, char *arg) {
 	std::string templ;
@@ -681,7 +781,7 @@ populate_target(std::string& target, char *arg) {
 
 	char *s = strdup(templ.c_str());
 
-	if (mkdtemp(s) != NULL) {
+	if (s != NULL && mkdtemp(s) != NULL) {
 		char *x = realpath(s, NULL);
 
 		if (x)
@@ -757,7 +857,11 @@ populate_metadata(libtorrent::add_torrent_params& p, const char *arg) {
 				ec.message().c_str()), false);
 
 		if (params.browse_only)
+#if LIBTORRENT_VERSION_NUM < 10200
 			p.flags |= libtorrent::add_torrent_params::flag_paused;
+#else
+			p.flags |= libtorrent::torrent_flags::paused;
+#endif
 	} else if (uri.find("magnet:") == 0) {
 		libtorrent::error_code ec;
 
@@ -791,7 +895,11 @@ populate_metadata(libtorrent::add_torrent_params& p, const char *arg) {
 				ec.message().c_str()), false);
 
 		if (params.browse_only)
+#if LIBTORRENT_VERSION_NUM < 10200
 			p.flags |= libtorrent::add_torrent_params::flag_paused;
+#else
+			p.flags |= libtorrent::torrent_flags::paused;
+#endif
 	}
 
 	return true;
@@ -860,6 +968,8 @@ main(int argc, char *argv[]) {
 	btfs_ops.open = btfs_open;
 	btfs_ops.read = btfs_read;
 	btfs_ops.statfs = btfs_statfs;
+	btfs_ops.listxattr = btfs_listxattr;
+	btfs_ops.getxattr = btfs_getxattr;
 	btfs_ops.init = btfs_init;
 	btfs_ops.destroy = btfs_destroy;
 
@@ -917,8 +1027,13 @@ main(int argc, char *argv[]) {
 
 	libtorrent::add_torrent_params p;
 
+#if LIBTORRENT_VERSION_NUM < 10200
 	p.flags &= ~libtorrent::add_torrent_params::flag_auto_managed;
 	p.flags &= ~libtorrent::add_torrent_params::flag_paused;
+#else
+	p.flags &= ~libtorrent::torrent_flags::auto_managed;
+	p.flags &= ~libtorrent::torrent_flags::paused;
+#endif
 	p.save_path = target + "/files";
 
 	if (mkdir(p.save_path.c_str(), 0777) < 0)
